"""
psbook command.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import argparse
import sys
import warnings
from typing import List

from psutils.argparse import (
    HelpFormatter,
    PaperContext,
    add_basic_arguments,
    parserange,
    parsespecs,
)
from psutils.transformers import file_transform
from psutils.warnings import die, simple_warning


def get_parser() -> argparse.ArgumentParser:
    # Command-line arguments
    parser = argparse.ArgumentParser(
        description="Rearrange pages in a PDF or PostScript document into signatures.",
        formatter_class=HelpFormatter,
        usage="%(prog)s [OPTION...] [INFILE [OUTFILE]]",
        add_help=False,
        epilog="""
PAGES is a comma-separated list of pages and page ranges; see pstops(1)
for more details.
""",
    )
    warnings.showwarning = simple_warning(parser.prog)

    # Command-line parser
    parser.add_argument(
        "-s",
        "--signature",
        type=int,
        default=0,
        help="""\
number of pages per signature;
0 = all pages in one signature [default];
1 = do not rearrange the pages;
otherwise, a multiple of 4""",
    )
    add_basic_arguments(parser)

    return parser


# pylint: disable=dangerous-default-value
def psbook(argv: List[str] = sys.argv[1:]) -> None:
    args = get_parser().parse_intermixed_args(argv)

    if args.signature > 1 and args.signature % 4 != 0:
        die("signature must be a multiple of 4")

    # Get number of pages
    paper_context = PaperContext()
    specs, modulo, flipping = parsespecs("0", paper_context)
    with file_transform(
        args.infile, args.outfile, None, None, specs, 0, False
    ) as transform:
        input_pages = transform.pages()

        def page_index_to_real_page(signature: int, page_number: int) -> int:
            real_page = page_number - page_number % signature
            page_on_sheet = page_number % 4
            recto_verso = (page_number % signature) // 2
            if page_on_sheet in (0, 3):
                real_page += signature - 1 - recto_verso
            else:
                real_page += recto_verso
            return real_page + 1

        # Adjust for signature size
        signature = args.signature
        if signature == 0:
            maxpage = input_pages + (4 - input_pages % 4) % 4
            signature = maxpage
        else:
            maxpage = input_pages + (signature - input_pages % signature) % signature

        # Compute page list
        page_list = []
        for page in range(maxpage):
            real_page = page_index_to_real_page(signature, page)
            page_list.append(str(real_page) if real_page <= input_pages else "_")

        # Rearrange pages
        transform.transform_pages(
            parserange(",".join(page_list)),
            flipping,
            False,
            False,
            False,
            modulo,
            args.verbose,
        )


if __name__ == "__main__":
    psbook()



"""
PSUtils basic types.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from dataclasses import dataclass
from typing import List, NamedTuple

from .warnings import die


@dataclass
class Rectangle:
    width: float
    height: float

    def __str__(self) -> str:
        return f"{self.width:n}x{self.height:n} pt"


@dataclass
class Range:
    start: int
    end: int
    text: str


Offset = NamedTuple("Offset", [("x", float), ("y", float)])


@dataclass
class PageSpec:
    reversed: bool = False
    pageno: int = 0
    rotate: int = 0
    hflip: bool = False
    vflip: bool = False
    scale: float = 1.0
    off: Offset = Offset(0.0, 0.0)

    def has_transform(self) -> bool:
        return (
            self.rotate != 0
            or self.hflip
            or self.vflip
            or self.scale != 1.0
            or self.off != Offset(0.0, 0.0)
        )


class PageList:
    def __init__(
        self,
        total_pages: int,
        pagerange: List[Range],
        reverse: bool,
        odd: bool,
        even: bool,
    ) -> None:
        self.pages: List[int] = []
        for range_ in pagerange:
            inc = -1 if range_.end < range_.start else 1
            currentpg = range_.start
            while range_.end - currentpg != -inc:
                if currentpg > total_pages:
                    die(f"page range {range_.text} is invalid", 2)
                if not (odd and (not even) and currentpg % 2 == 0) and not (
                    even and not odd and currentpg % 2 == 1
                ):
                    self.pages.append(currentpg - 1)
                currentpg += inc
        if reverse:
            self.pages.reverse()

    # Returns -1 for an inserted blank page (page number '_')
    def real_page(self, pagenum: int) -> int:
        try:
            return self.pages[pagenum]
        except IndexError:
            return 0

    def num_pages(self) -> int:
        return len(self.pages)



"""
epsffit tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path
from itertools import combinations
from typing import Callable, List, Iterable

from pytest import CaptureFixture

from testutils import file_test, make_tests, Case
from psutils.command.epsffit import epsffit

OPTIONS = ["aspect", "center", "maximize", "rotate"]

OPTION_SETS = [
    subset for n in range(len(OPTIONS) + 1) for subset in combinations(OPTIONS, n)
]


def argify(options: Iterable[str]) -> Iterable[str]:
    return [f"-{opt[0]}" for opt in options]


option_cases = []
for opts in OPTION_SETS:
    option_cases.append(
        Case(
            "-".join(opts) or "no-options",
            ["100pt", "100pt", "200pt", "300pt", *argify(opts)],
            "tiger.eps",
        )
    )

pytestmark = make_tests(
    epsffit,
    Path(__file__).parent.resolve() / "test-files",
    *option_cases,
    Case(
        "showpage",
        ["-s", "100pt", "100pt", "200pt", "300pt"],
        "tiger.eps",
    ),
    Case(
        "h-texlive",
        ["-c", "0", "0", "600", "368"],
        "plot.eps",
    ),
    Case(
        "m-texlive",
        ["-m", "0", "0", "368", "500"],
        "plot.eps",
    ),
    Case(
        "v-texlive",
        ["-c", "0", "0", "500", "400"],
        "plot.eps",
    ),
)


# pylint: disable=similarities
def test_epsffit(
    function: Callable[[List[str]], None],
    case: Case,
    fixture_dir: Path,
    capsys: CaptureFixture[str],
    datafiles: Path,
    regenerate_input: bool,
    regenerate_expected: bool,
) -> None:
    file_test(
        function,
        case,
        fixture_dir,
        capsys,
        datafiles,
        ".eps",
        regenerate_input,
        regenerate_expected,
    )



"""
PSUtils document reader classes.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path
import re
from typing import List, Tuple, Union, Type, IO

from pypdf import PdfReader as PdfReaderBase
from pypdf._utils import StrByteType

from .types import Rectangle
from .warnings import die


class PdfReader(PdfReaderBase):
    def __init__(
        self,
        stream: Union[StrByteType, Path],
        strict: bool = False,
        password: Union[str, bytes, None] = None,
    ) -> None:
        super().__init__(stream, strict, password)
        assert len(self.pages) > 0
        mediabox = self.pages[0].mediabox
        self.size = Rectangle(mediabox.width, mediabox.height)
        self.size_guessed = False


size_keywords = (
    b"DocumentMedia",
    b"PageBoundingBox",
    b"HiResBoundingBox",
    b"BoundingBox",
)


# FIXME: Store lists of lines, not file offsets.
class PsReader:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    def __init__(self, infile: IO[bytes]) -> None:
        self.infile = infile
        self.headerpos: int = 0
        self.pagescmt: int = 0
        self.endsetup: int = 0
        self.procset_pos: range = range(0, 0)  # pstops procset location
        self.num_pages: int = 0
        self.sizeheaders: List[int] = []
        self.pageptr: List[int] = []
        self.size = None
        self.size_guessed = False

        nesting = 0
        self.infile.seek(0)
        record, next_record, buffer = 0, 0, None
        file_sizes = {}
        for buffer in self.infile:
            next_record += len(buffer)
            if buffer.startswith(b"%%"):
                keyword, value = self.comment(buffer)
                if keyword is not None:
                    # If input paper size is not set, try to read it
                    if (
                        self.headerpos == 0
                        and self.size is None
                        and keyword in size_keywords
                    ):
                        assert value is not None
                        words = value.split(b" ")
                        if keyword == b"DocumentMedia" and len(words) > 2:
                            w = words[1].decode("utf-8", "ignore")
                            h = words[2].decode("utf-8", "ignore")
                            try:
                                file_sizes[keyword] = Rectangle(float(w), float(h))
                            except ValueError:
                                pass
                        elif len(words) == 4:
                            llx = words[0].decode("utf-8", "ignore")
                            lly = words[1].decode("utf-8", "ignore")
                            urx = words[2].decode("utf-8", "ignore")
                            ury = words[3].decode("utf-8", "ignore")
                            try:
                                file_sizes[keyword] = Rectangle(
                                    float(urx) - float(llx), float(ury) - float(lly)
                                )
                            except ValueError:
                                pass
                    if nesting == 0 and keyword == b"Page":
                        self.pageptr.append(record)
                    elif self.headerpos == 0 and (
                        keyword in size_keywords or keyword == b"DocumentPaperSizes"
                    ):
                        self.sizeheaders.append(record)
                    elif self.headerpos == 0 and keyword == b"Pages":
                        self.pagescmt = record
                    elif self.headerpos == 0 and keyword == b"EndComments":
                        self.headerpos = next_record
                    elif keyword in [
                        b"BeginDocument",
                        b"BeginBinary",
                        b"Begself.infile",
                    ]:
                        nesting += 1
                    elif keyword in [b"EndDocument", b"EndBinary", b"EndFile"]:
                        nesting -= 1
                    elif nesting == 0 and keyword == b"EndSetup":
                        self.endsetup = record
                    elif nesting == 0 and keyword == b"BeginProlog":
                        self.headerpos = next_record
                    elif nesting == 0 and buffer == b"%%BeginProcSet: PStoPS":
                        self.procset_pos = range(record, 0)
                    elif (
                        self.procset_pos.start > 0
                        and self.procset_pos.stop == 0
                        and keyword == b"EndProcSet"
                    ):
                        self.procset_pos = range(self.procset_pos.start, next_record)
                    elif nesting == 0 and keyword in [b"Trailer", b"EOF"]:
                        break
            elif self.headerpos == 0:
                self.headerpos = record
            record = next_record

        # If paper size was not already set, and we found a possible size in
        # the file, use it.
        if self.size is None and len(file_sizes) > 0:
            for keyword in size_keywords:
                file_size = file_sizes.get(keyword)
                if (
                    file_size is not None
                    and file_size.width != 0
                    and file_size.height != 0
                ):
                    self.size = file_size
                    if keyword in (b"BoundingBox", b"HiResBoundingBox"):
                        self.size_guessed = True
                    break
        self.num_pages = len(self.pageptr)
        self.pageptr.append(record)
        if self.endsetup == 0 or self.endsetup > self.pageptr[0]:
            self.endsetup = self.pageptr[0]

    # Return comment keyword and value if `line' is a DSC comment
    def comment(self, line: bytes) -> Union[Tuple[bytes, bytes], Tuple[None, None]]:
        m = re.match(b"%%([^:]+):?\\s+?(.*\\S?)\\s*$", line)
        return (m[1], m[2]) if m else (None, None)


def document_reader(file: IO[bytes], file_type: str) -> Union[PdfReader, PsReader]:
    constructor: Union[Type[PdfReader], Type[PsReader]]
    if file_type in (".ps", ".eps"):
        constructor = PsReader
    elif file_type == ".pdf":
        constructor = PdfReader
    else:
        die(f"incompatible file type `{file_type}'")
    return constructor(file)



"""
psbook tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path

from testutils import file_test, make_tests, Case, GeneratedInput
from psutils.command.psbook import psbook

pytestmark = make_tests(
    psbook,
    Path(__file__).parent.resolve() / "test-files",
    Case(
        "3",
        [],
        GeneratedInput("a4", 3),
    ),
    Case(
        "3-signature-4",
        ["-s", "4"],
        GeneratedInput("a4", 3),
    ),
    Case(
        "20",
        [],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-signature-4",
        ["-s", "4"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "invalid-signature-size",
        ["-s", "3"],
        "no-input",
        1,
    ),
    Case(
        "texlive",
        ["-s4"],
        GeneratedInput("a4", 11),
    ),
)
test_psbook = file_test



"""
Test configuration.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from typing import Any

from pytest import FixtureRequest, Parser, fixture


def pytest_generate_tests(metafunc: Any) -> None:
    if "file_type" in metafunc.fixturenames:
        metafunc.parametrize("file_type", [".ps", ".pdf"])


def pytest_addoption(parser: Parser) -> None:
    parser.addoption(
        "--regenerate-expected",
        action="store_true",
        help="regenerate the expected outputs",
    )
    parser.addoption(
        "--regenerate-input",
        action="store_true",
        help="regenerate the inputs",
    )


@fixture
def regenerate_expected(request: FixtureRequest) -> bool:
    opt = request.config.getoption("--regenerate-expected")
    assert isinstance(opt, bool)
    return opt


@fixture
def regenerate_input(request: FixtureRequest) -> bool:
    opt = request.config.getoption("--regenerate-input")
    assert isinstance(opt, bool)
    return opt



"""
psresize tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path

from testutils import file_test, make_tests, Case, GeneratedInput
from psutils.command.psresize import psresize

pytestmark = make_tests(
    psresize,
    Path(__file__).parent.resolve() / "test-files",
    Case(
        "20-A4",
        ["-p", "a4"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-A3",
        ["-p", "a3"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-A5",
        ["-p", "a5"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-Letter",
        ["-p", "letter"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-A5in-A4",
        ["-p", "a4"],
        GeneratedInput("a5", 20),
    ),
    Case(
        "20-A3in-A4",
        ["-p", "a4"],
        GeneratedInput("a3", 20),
    ),
)
test_psresize = file_test



"""
psjoin tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from contextlib import redirect_stdout
from pathlib import Path
from typing import List

from testutils import make_tests, file_test, Case, GeneratedInput
from psutils.command.psjoin import psjoin

FIXTURE_DIR = Path(__file__).parent.resolve() / "test-files"


def psjoin_to_file(args: List[str]) -> None:
    output_file = args.pop()
    args.append(args[-1])
    with open(output_file, "w", encoding="utf-8") as f:
        with redirect_stdout(f):
            psjoin(args)


pytestmark = make_tests(
    psjoin_to_file,
    Path(__file__).parent.resolve() / "test-files",
    Case(
        "1-2",
        [],
        GeneratedInput("a4", 1),
    ),
    Case(
        "1-2-even",
        ["--even"],
        GeneratedInput("a4", 1),
    ),
    Case(
        "1-2-nostrip",
        ["--nostrip"],
        GeneratedInput("a4", 1),
    ),
    Case(
        "1-2-save",
        ["--save"],
        GeneratedInput("a4", 1),
    ),
)
test_psjoin = file_test



"""
PSUtils document transformer classes.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import io
import sys
import shutil
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import List, Optional, Union, Iterator, IO
from warnings import warn

from pypdf import PdfWriter, Transformation
from pypdf.annotations import PolyLine

from .argparse import parserange
from .io import setup_input_and_output
from .readers import PsReader, PdfReader, document_reader
from .types import Rectangle, Range, Offset, PageSpec, PageList
from .warnings import die


def page_index_to_page_number(
    spec: PageSpec, maxpage: int, modulo: int, pagebase: int
) -> int:
    return (maxpage - pagebase - modulo if spec.reversed else pagebase) + spec.pageno


class DocumentTransform(ABC):
    def __init__(self) -> None:
        self.in_size: Optional[Rectangle]
        self.specs: List[List[PageSpec]]

    @abstractmethod
    def pages(self) -> int:
        pass

    @abstractmethod
    def write_header(self, maxpage: int, modulo: int) -> None:
        pass

    @abstractmethod
    def write_page_comment(self, pagelabel: str, outputpage: int) -> None:
        pass

    @abstractmethod
    def write_page(
        self,
        page_list: PageList,
        outputpage: int,
        page_specs: List[PageSpec],
        maxpage: int,
        modulo: int,
        pagebase: int,
    ) -> None:
        pass

    @abstractmethod
    def finalize(self) -> None:
        pass

    def transform_pages(
        self,
        pagerange: Optional[List[Range]],
        flipping: bool,
        reverse: bool,
        odd: bool,
        even: bool,
        modulo: int,
        verbose: bool,
    ) -> None:
        if self.in_size is None and flipping:
            die("input page size must be set when flipping the page")

        # Page spec routines for page rearrangement
        def abs_page(n: int) -> int:
            if n < 0:
                n += self.pages() + 1
                n = max(n, 1)
            return n

        def transform_pages(
            pagerange: Optional[List[Range]], odd: bool, even: bool, reverse: bool
        ) -> None:
            outputpage = 0
            # If no page range given, select all pages
            if pagerange is None:
                pagerange = parserange("1-_1")

            # Normalize end-relative pageranges
            for range_ in pagerange:
                range_.start = abs_page(range_.start)
                range_.end = abs_page(range_.end)

            # Get list of pages
            page_list = PageList(self.pages(), pagerange, reverse, odd, even)

            # Calculate highest page number output (including any blanks)
            maxpage = (
                page_list.num_pages()
                + (modulo - page_list.num_pages() % modulo) % modulo
            )

            # Rearrange pages
            self.write_header(maxpage, modulo)
            pagebase = 0
            while pagebase < maxpage:
                for page in self.specs:
                    # Construct the page label from the input page numbers
                    pagelabels = []
                    for spec in page:
                        n = page_list.real_page(
                            page_index_to_page_number(spec, maxpage, modulo, pagebase)
                        )
                        pagelabels.append(str(n + 1) if n >= 0 else "*")
                    pagelabel = ",".join(pagelabels)
                    outputpage += 1
                    self.write_page_comment(pagelabel, outputpage)
                    if verbose:
                        sys.stderr.write(f"[{pagelabel}] ")
                    self.write_page(
                        page_list, outputpage, page, maxpage, modulo, pagebase
                    )

                pagebase += modulo

            self.finalize()
            if verbose:
                print(f"\nWrote {outputpage} pages", file=sys.stderr)

        # Output the pages
        transform_pages(pagerange, odd, even, reverse)


# FIXME: Extract PsWriter.
class PsTransform(DocumentTransform):  # pylint: disable=too-many-instance-attributes
    # PStoPS procset
    # Wrap showpage, erasepage and copypage in our own versions.
    # Nullify paper size operators.
    procset = """userdict begin
[/showpage/erasepage/copypage]{dup where{pop dup load
 type/operatortype eq{ /PStoPSenablepage cvx 1 index
 load 1 array astore cvx {} bind /ifelse cvx 4 array
 astore cvx def}{pop}ifelse}{pop}ifelse}forall
 /PStoPSenablepage true def
[/letter/legal/executivepage/a4/a4small/b5/com10envelope
 /monarchenvelope/c5envelope/dlenvelope/lettersmall/note
 /folio/quarto/a5]{dup where{dup wcheck{exch{}put}
 {pop{}def}ifelse}{pop}ifelse}forall
/setpagedevice {pop}bind 1 index where{dup wcheck{3 1 roll put}
 {pop def}ifelse}{def}ifelse
/PStoPSmatrix matrix currentmatrix def
/PStoPSxform matrix def/PStoPSclip{clippath}def
/defaultmatrix{PStoPSmatrix exch PStoPSxform exch concatmatrix}bind def
/initmatrix{matrix defaultmatrix setmatrix}bind def
/initclip[{matrix currentmatrix PStoPSmatrix setmatrix
 [{currentpoint}stopped{$error/newerror false put{newpath}}
 {/newpath cvx 3 1 roll/moveto cvx 4 array astore cvx}ifelse]
 {[/newpath cvx{/moveto cvx}{/lineto cvx}
 {/curveto cvx}{/closepath cvx}pathforall]cvx exch pop}
 stopped{$error/errorname get/invalidaccess eq{cleartomark
 $error/newerror false put cvx exec}{stop}ifelse}if}bind aload pop
 /initclip dup load dup type dup/operatortype eq{pop exch pop}
 {dup/arraytype eq exch/packedarraytype eq or
  {dup xcheck{exch pop aload pop}{pop cvx}ifelse}
  {pop cvx}ifelse}ifelse
 {newpath PStoPSclip clip newpath exec setmatrix} bind aload pop]cvx def
/initgraphics{initmatrix newpath initclip 1 setlinewidth
 0 setlinecap 0 setlinejoin []0 setdash 0 setgray
 10 setmiterlimit}bind def
end"""

    def __init__(
        self,
        reader: PsReader,
        outfile: IO[bytes],
        size: Optional[Rectangle],
        in_size: Optional[Rectangle],
        specs: List[List[PageSpec]],
        draw: float,
        in_size_guessed: bool,
    ):
        super().__init__()
        self.reader = reader
        self.outfile = outfile
        self.draw = draw
        self.specs = specs
        self.in_size_guessed = in_size_guessed

        self.use_procset = any(
            len(page) > 1 or page[0].has_transform() for page in specs
        )

        self.size = size
        if in_size is None:
            if reader.size is not None:
                in_size = reader.size
            elif size is not None:
                in_size = size
        self.in_size = in_size

    def pages(self) -> int:
        return self.reader.num_pages

    def write_header(self, maxpage: int, modulo: int) -> None:
        # FIXME: doesn't cope properly with loaded definitions
        ignorelist = [] if self.size is None else self.reader.sizeheaders
        self.reader.infile.seek(0)
        if self.reader.pagescmt:
            self.fcopy(self.reader.pagescmt, ignorelist)
            try:
                _ = self.reader.infile.readline()
            except IOError:
                die("I/O error in header", 2)
            if self.size is not None:
                if self.in_size_guessed:
                    warn(f"required input paper size was guessed as {self.in_size}")
                self.write(
                    f"%%DocumentMedia: plain {int(self.size.width)} {int(self.size.height)} 0 () ()"
                )
                self.write(
                    f"%%BoundingBox: 0 0 {int(self.size.width)} {int(self.size.height)}"
                )
            pagesperspec = len(self.specs)
            self.write(f"%%Pages: {int(maxpage / modulo) * pagesperspec} 0")
        self.fcopy(self.reader.headerpos, ignorelist)
        if self.use_procset:
            self.write(f"%%BeginProcSet: PStoPS 1 15\n{self.procset}")
            self.write("%%EndProcSet")

        # Write prologue to end of setup section, skipping our procset if present
        # and we're outputting it (this allows us to upgrade our procset)
        if self.reader.procset_pos and self.use_procset:
            self.fcopy(self.reader.procset_pos.start, [])
            self.reader.infile.seek(self.reader.procset_pos.stop)
        self.fcopy(self.reader.endsetup, [])

        # Save transformation from original to current matrix
        if not self.reader.procset_pos and self.use_procset:
            self.write(
                """userdict/PStoPSxform PStoPSmatrix matrix currentmatrix
 matrix invertmatrix matrix concatmatrix
 matrix invertmatrix put"""
            )

        # Write from end of setup to start of pages
        self.fcopy(self.reader.pageptr[0], [])

    def write(self, text: str) -> None:
        self.outfile.write((text + "\n").encode("utf-8"))

    def write_page_comment(self, pagelabel: str, outputpage: int) -> None:
        self.write(f"%%Page: ({pagelabel}) {outputpage}")

    def write_page(
        self,
        page_list: PageList,
        outputpage: int,
        page_specs: List[PageSpec],
        maxpage: int,
        modulo: int,
        pagebase: int,
    ) -> None:
        spec_page_number = 0
        for spec in page_specs:
            page_number = page_index_to_page_number(spec, maxpage, modulo, pagebase)
            real_page = page_list.real_page(page_number)
            if page_number < page_list.num_pages() and 0 <= real_page < self.pages():
                # Seek the page
                pagenum = real_page
                self.reader.infile.seek(self.reader.pageptr[pagenum])
                try:
                    line = self.reader.infile.readline()
                    keyword, _ = self.reader.comment(line)
                    assert keyword == b"Page"
                except IOError:
                    die(f"I/O error seeking page {pagenum}", 2)
            if self.use_procset:
                self.write("userdict/PStoPSsaved save put")
            if spec.has_transform():
                self.write("PStoPSmatrix setmatrix")
                if spec.off != Offset(0.0, 0.0):
                    self.write(f"{spec.off.x:f} {spec.off.y:f} translate")
                if spec.rotate != 0:
                    self.write(f"{spec.rotate % 360} rotate")
                if spec.hflip == 1:
                    assert self.in_size is not None
                    self.write(
                        f"[ -1 0 0 1 {self.in_size.width * spec.scale:g} 0 ] concat"
                    )
                if spec.vflip == 1:
                    assert self.in_size is not None
                    self.write(
                        f"[ 1 0 0 -1 0 {self.in_size.height * spec.scale:g} ] concat"
                    )
                if spec.scale != 1.0:
                    self.write(f"{spec.scale:f} dup scale")
                self.write("userdict/PStoPSmatrix matrix currentmatrix put")
                if self.in_size is not None:
                    w, h = self.in_size.width, self.in_size.height
                    self.write(
                        f"""userdict/PStoPSclip{{0 0 moveto
 {w:f} 0 rlineto 0 {h:f} rlineto {-w:f} 0 rlineto
 closepath}}put initclip"""
                    )
                    if self.draw > 0:
                        self.write(
                            f"gsave clippath 0 setgray {self.draw} setlinewidth stroke grestore"
                        )
            if spec_page_number < len(page_specs) - 1:
                self.write("/PStoPSenablepage false def")
            if (
                self.reader.procset_pos
                and page_number < page_list.num_pages()
                and real_page < self.pages()
            ):
                # Search for page setup
                while True:
                    try:
                        line = self.reader.infile.readline()
                    except IOError:
                        die(f"I/O error reading page setup {outputpage}", 2)
                    if line.startswith(b"PStoPSxform"):
                        break
                    try:
                        self.write(line.decode())
                    except IOError:
                        die(f"I/O error writing page setup {outputpage}", 2)
            if not self.reader.procset_pos and self.use_procset:
                self.write("PStoPSxform concat")
            if page_number < page_list.num_pages() and 0 <= real_page < self.pages():
                # Write the body of a page
                self.fcopy(self.reader.pageptr[real_page + 1], [])
            else:
                self.write("showpage")
            if self.use_procset:
                self.write("PStoPSsaved restore")
            spec_page_number += 1

    def finalize(self) -> None:
        # Write trailer
        self.reader.infile.seek(self.reader.pageptr[self.pages()])
        shutil.copyfileobj(self.reader.infile, self.outfile)
        self.outfile.flush()

    # Copy input file from current position up to new position to output file,
    # ignoring the lines starting at something ignorelist points to.
    # Updates ignorelist.
    def fcopy(self, upto: int, ignorelist: List[int]) -> None:
        here = self.reader.infile.tell()
        while len(ignorelist) > 0 and ignorelist[0] < upto:
            while len(ignorelist) > 0 and ignorelist[0] < here:
                ignorelist.pop(0)
            if len(ignorelist) > 0:
                self.fcopy(ignorelist[0], [])
            try:
                self.reader.infile.readline()
            except IOError:
                die("I/O error", 2)
            ignorelist.pop(0)
            here = self.reader.infile.tell()

        try:
            self.outfile.write(self.reader.infile.read(upto - here))
        except IOError:
            die("I/O error", 2)


class PdfTransform(DocumentTransform):
    def __init__(
        self,
        reader: PdfReader,
        outfile: IO[bytes],
        size: Optional[Rectangle],
        in_size: Optional[Rectangle],
        specs: List[List[PageSpec]],
        draw: float,
    ):
        super().__init__()
        self.outfile = outfile
        self.reader = reader
        self.writer = PdfWriter()
        self.draw = draw
        self.specs = specs

        if in_size is None:
            in_size = reader.size
        if size is None:
            size = in_size

        self.size = size
        self.in_size = in_size

    def pages(self) -> int:
        return len(self.reader.pages)

    def write_header(self, maxpage: int, modulo: int) -> None:
        pass

    def write_page_comment(self, pagelabel: str, outputpage: int) -> None:
        pass

    def write_page(
        self,
        page_list: PageList,
        outputpage: int,
        page_specs: List[PageSpec],
        maxpage: int,
        modulo: int,
        pagebase: int,
    ) -> None:
        assert self.in_size
        page_number = page_index_to_page_number(
            page_specs[0], maxpage, modulo, pagebase
        )
        real_page = page_list.real_page(page_number)
        if (  # pylint: disable=too-many-boolean-expressions
            len(page_specs) == 1
            and not page_specs[0].has_transform()
            and page_number < page_list.num_pages()
            and 0 <= real_page < len(self.reader.pages)
            and self.draw == 0
            and (
                self.in_size.width is None
                or (
                    self.in_size.width == self.reader.pages[real_page].mediabox.width
                    and self.in_size.height
                    == self.reader.pages[real_page].mediabox.height
                )
            )
        ):
            self.writer.add_page(self.reader.pages[real_page])
        else:
            # Add a blank page of the correct size to the end of the document
            outpdf_page = self.writer.add_blank_page(self.size.width, self.size.height)
            for spec in page_specs:
                page_number = page_index_to_page_number(spec, maxpage, modulo, pagebase)
                real_page = page_list.real_page(page_number)
                if page_number < page_list.num_pages() and 0 <= real_page < len(
                    self.reader.pages
                ):
                    # Calculate input page transformation
                    t = Transformation()
                    if spec.hflip:
                        t = t.transform(
                            Transformation((-1, 0, 0, 1, self.in_size.width, 0))
                        )
                    elif spec.vflip:
                        t = t.transform(
                            Transformation((1, 0, 0, -1, 0, self.in_size.height))
                        )
                    if spec.rotate != 0:
                        t = t.rotate(spec.rotate % 360)
                    if spec.scale != 1.0:
                        t = t.scale(spec.scale, spec.scale)
                    if spec.off != Offset(0.0, 0.0):
                        t = t.translate(spec.off.x, spec.off.y)
                    # Merge input page into the output document
                    outpdf_page.merge_transformed_page(self.reader.pages[real_page], t)
                    if self.draw > 0:  # FIXME: draw the line at the requested width
                        mediabox = self.reader.pages[real_page].mediabox
                        line = PolyLine(
                            vertices=[
                                (
                                    mediabox.left + spec.off.x,
                                    mediabox.bottom + spec.off.y,
                                ),
                                (mediabox.left + spec.off.x, mediabox.top + spec.off.y),
                                (
                                    mediabox.right + spec.off.x,
                                    mediabox.top + spec.off.y,
                                ),
                                (
                                    mediabox.right + spec.off.x,
                                    mediabox.bottom + spec.off.y,
                                ),
                                (
                                    mediabox.left + spec.off.x,
                                    mediabox.bottom + spec.off.y,
                                ),
                            ],
                        )
                        self.writer.add_annotation(outpdf_page, line)

    def finalize(self) -> None:
        # PyPDF seeks, so write to a buffer first in case outfile is stdout.
        buf = io.BytesIO()
        self.writer.write(buf)
        buf.seek(0)
        self.outfile.write(buf.read())
        self.outfile.flush()


def document_transform(
    indoc: Union[PdfReader, PsReader],
    outfile: IO[bytes],
    size: Optional[Rectangle],
    in_size: Optional[Rectangle],
    specs: List[List[PageSpec]],
    draw: float,
    in_size_guessed: bool,
) -> Union[PdfTransform, PsTransform]:
    if isinstance(indoc, PsReader):
        return PsTransform(indoc, outfile, size, in_size, specs, draw, in_size_guessed)
    if isinstance(indoc, PdfReader):
        return PdfTransform(indoc, outfile, size, in_size, specs, draw)
    die("unknown document type")


@contextmanager
def file_transform(
    infile_name: str,
    outfile_name: str,
    size: Optional[Rectangle],
    in_size: Optional[Rectangle],
    specs: List[List[PageSpec]],
    draw: float,
    in_size_guessed: bool,
) -> Iterator[Union[PdfTransform, PsTransform]]:
    with setup_input_and_output(infile_name, outfile_name) as (
        infile,
        file_type,
        outfile,
    ):
        doc = document_reader(infile, file_type)
        yield document_transform(
            doc, outfile, size, in_size, specs, draw, in_size_guessed
        )



"""
pstops tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import os
from pathlib import Path
from unittest import mock

from testutils import file_test, make_tests, Case, GeneratedInput
from psutils.command.pstops import pstops

pytestmark = make_tests(
    pstops,
    Path(__file__).parent.resolve() / "test-files",
    # Test backwards-compatible specs syntax without --specs flag.
    Case(
        "offsets",
        ["0(100pt,200pt)"],
        GeneratedInput("a4", 1),
    ),
    Case(
        "negative-offsets",
        ["--specs", "0(-100pt,-200pt)"],
        GeneratedInput("a4", 1),
    ),
    Case(
        "correct-angles",
        ["-pa4", "--specs", "0L(1w,0)+0R(0,1h)"],
        GeneratedInput("a5", 1),
    ),
    Case(
        "multiple-pages",
        ["--specs", "2:0(100pt,200pt),1(-200pt,100pt)"],
        GeneratedInput("a4", 2),
    ),
    Case(
        "multiple-turns-and-flips",
        ["--specs", "0LLRVHVHV(700pt,0pt)"],
        GeneratedInput("a4", 1),
    ),
    Case(
        "invalid-pagespecs",
        ["--specs=foo"],
        GeneratedInput("a4", 1),
        1,
    ),
    Case(
        "texlive",
        ["-pa4", "--specs", "2:0L@.7(21cm,0)+1L@.7(21cm,14.85cm)"],
        GeneratedInput("a4", 11),
    ),
    Case(  # Test we can refer to the paper size in a dimension when output size is not set
        "default-paper-size",
        ["--specs", "0L@.7(1w,0)+0L@.7(1w,.5h)"],
        GeneratedInput("a4", 1),
    ),
    Case(
        "man-page-example",
        [
            "-S",
            "4:-3L@.7(1w,0h)+0L@.7(1w,0.5h),1L@.7(1w,0h)+-2L@.7(1w,0.5h)",
        ],
        GeneratedInput("a4", 20),
    ),
)
with mock.patch.dict(os.environ, {"PAPERSIZE": "A4"}):
    test_pstops = file_test



"""
PSUtils I/O utilities.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import io
import os
import sys
from contextlib import contextmanager
from typing import Tuple, Optional, Iterator, IO

import puremagic

from .warnings import die


@contextmanager
def setup_input_and_output(
    infile_name: Optional[str], outfile_name: Optional[str]
) -> Iterator[Tuple[IO[bytes], str, IO[bytes]]]:
    # Set up input
    infile: Optional[IO[bytes]] = None
    if infile_name is not None:
        try:
            infile = open(infile_name, "rb")
        except IOError:
            die(f"cannot open input file {infile_name}")
    else:
        infile = os.fdopen(sys.stdin.fileno(), "rb", closefd=False)

    # Find MIME type of input
    data = infile.read(16)
    file_type = puremagic.from_string(data)

    # Slurp infile into a seekable BytesIO
    infile = io.BytesIO(data + infile.read())

    # Set up output
    if outfile_name is not None:
        try:
            outfile = open(outfile_name, "wb")
        except IOError:
            die(f"cannot open output file {outfile_name}")
    else:
        outfile = os.fdopen(sys.stdout.fileno(), "wb", closefd=False)

    # Context manager
    try:
        yield infile, file_type, outfile
    finally:
        infile.close()
        outfile.close()



"""
pdfnup tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path
from typing import List, Callable

from pytest import CaptureFixture

from testutils import file_test, make_tests, Case
from psutils.command.psnup import psnup

pytestmark = make_tests(
    psnup,
    Path(__file__).parent.resolve() / "test-files",
    Case(
        "recursive-links",
        ["-2"],
        "recursive-links",
    ),
)


def test_pdfnup(
    function: Callable[[List[str]], None],
    case: Case,
    fixture_dir: Path,
    capsys: CaptureFixture[str],
    datafiles: Path,
    regenerate_input: bool,
    regenerate_expected: bool,
) -> None:
    file_test(
        function,
        case,
        fixture_dir,
        capsys,
        datafiles,
        ".pdf",
        regenerate_input,
        regenerate_expected,
    )



"""
psutils tests utility routines.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import os
import sys
import subprocess
import re
import difflib
import shutil
from contextlib import ExitStack
from pathlib import Path
from dataclasses import dataclass
from unittest.mock import patch
from typing import Any, Callable, List, Iterator, Optional, Union
from warnings import warn

import pytest
from pytest import CaptureFixture, mark, param
from wand.image import Image  # type: ignore

if sys.version_info[:2] >= (3, 11):
    from contextlib import chdir
else:
    from contextlib import contextmanager

    @contextmanager
    def chdir(path: os.PathLike[str]) -> Iterator[None]:
        old_dir = os.getcwd()
        os.chdir(path)
        try:
            yield
        finally:
            os.chdir(old_dir)


@dataclass
class GeneratedInput:
    paper: str
    pages: int
    border: int = 1


@dataclass
class Case:
    name: str
    args: List[str]
    input: Union[GeneratedInput, str]
    error: Optional[int] = None


def remove_creation_date(lines: List[str]) -> List[str]:
    return [l for l in lines if not re.match(r"(% )?%%CreationDate", l)]


def compare_text_files(
    capsys: CaptureFixture[str],
    output_file: os.PathLike[str],
    expected_file: os.PathLike[str],
) -> bool:
    with ExitStack() as stack:
        out_fd = stack.enter_context(open(output_file, encoding="ascii"))
        exp_fd = stack.enter_context(open(expected_file, encoding="ascii"))
        output_lines = remove_creation_date(out_fd.readlines())
        expected_lines = remove_creation_date(exp_fd.readlines())
        diff = list(
            difflib.unified_diff(
                output_lines, expected_lines, str(output_file), str(expected_file)
            )
        )
        if len(diff) > 0:
            with capsys.disabled():
                sys.stdout.writelines(diff)
        return len(diff) == 0
    return False


def image_to_bytes(image_file: os.PathLike[str]) -> bytes:
    with Image(filename=image_file) as image:
        bytestr = b""
        # FIXME: If comparison fails, save images that differ for debugging
        for i, frame in enumerate(image.sequence):  # pylint: disable=unused-variable
            frame_img = Image(image=frame)
            bytestr += frame_img.make_blob("pnm")
            # frame_img.save(filename=f"{image_file}-{i}.pnm")
        return bytestr
    return b""


def compare_image_files(
    capsys: CaptureFixture[str],
    output_file: os.PathLike[str],
    expected_file: os.PathLike[str],
) -> bool:
    with ExitStack() as stack:
        out_fd = stack.enter_context(open(output_file, "rb"))
        exp_fd = stack.enter_context(open(expected_file, "rb"))
        output = out_fd.read()
        expected = exp_fd.read()
        if output == expected:
            return True
        output_bytes = image_to_bytes(output_file)
        expected_bytes = image_to_bytes(expected_file)
        if output_bytes == expected_bytes:
            with capsys.disabled():
                warn(
                    f"{output_file} not identical to {expected_file} but looks the same"
                )
            return True
    return False


def compare_strings(
    capsys: CaptureFixture[str],
    output: str,
    output_file: os.PathLike[str],
    expected_file: os.PathLike[str],
) -> bool:
    with open(output_file, "w", encoding="ascii") as f:
        f.write(output)
    return compare_text_files(capsys, output_file, expected_file)


def file_test(
    function: Callable[[List[str]], None],
    case: Case,
    fixture_dir: Path,
    capsys: CaptureFixture[str],
    datafiles: Path,
    file_type: str,
    regenerate_input: bool,
    regenerate_expected: bool,
) -> None:
    module_name = function.__name__
    expected_file = fixture_dir / module_name / case.name / "expected"
    expected_stderr = (
        fixture_dir / module_name / case.name / f"expected-stderr-{file_type[1:]}.txt"
    )
    if not os.path.exists(expected_stderr):
        expected_stderr = fixture_dir / module_name / case.name / "expected-stderr.txt"
    if isinstance(case.input, str):
        test_file = fixture_dir / case.input
    else:
        basename = f"{case.input.paper}-{case.input.pages}"
        if case.input.border != 1:
            basename += f"-{case.input.border}"
        test_file = fixture_dir / basename
    if regenerate_input and isinstance(case.input, GeneratedInput):
        make_test_input(
            case.input.paper, case.input.pages, test_file, case.input.border
        )
    output_file = datafiles / "output"
    full_args = [*case.args, str(test_file.with_suffix(file_type)), str(output_file)]
    patched_argv = [module_name, *(sys.argv[1:])]
    with chdir(datafiles):
        correct_output = True
        if case.error is None:
            with patch("sys.argv", patched_argv):
                function(full_args)
            if regenerate_expected:
                shutil.copyfile(output_file, expected_file.with_suffix(file_type))
                correct_output = True
            else:
                comparer = (
                    compare_text_files
                    if file_type in (".ps", ".eps")
                    else compare_image_files
                )
                correct_output = comparer(
                    capsys, output_file, expected_file.with_suffix(file_type)
                )
        else:
            with pytest.raises(SystemExit) as e:
                with patch("sys.argv", patched_argv):
                    function(full_args)
            assert e.value.code == case.error
        if regenerate_expected:
            with open(expected_stderr, "w", encoding="utf-8") as f:
                f.write(capsys.readouterr().err)
            correct_stderr = True
        else:
            correct_stderr = compare_strings(
                capsys,
                capsys.readouterr().err,
                datafiles / "stderr.txt",
                expected_stderr,
            )
        if not (correct_output and correct_stderr):
            bad_results: List[str] = []
            if not correct_output:
                bad_results.append("output")
            if not correct_stderr:
                bad_results.append("stderr")
            raise ValueError(
                f"test {','.join(bad_results)} does not match expected output"
            )


def make_tests(
    function: Callable[..., Any],
    fixture_dir: Path,
    *tests: Case,
) -> Any:
    ids = []
    test_cases = []
    for t in tests:
        ids.append(t.name)
        test_cases.append(t)
    return mark.parametrize(
        "function,case,fixture_dir",
        [
            param(
                function,
                case,
                fixture_dir,
                marks=mark.datafiles,
            )
            for case in test_cases
        ],
        ids=ids,
    )


# Make a test PostScript or PDF file of a given number of pages
# Requires a2ps and ps2pdf
# Simply writes a large page number on each page
def make_test_input(
    paper: str, pages: int, file: Path, border: Optional[int] = 1
) -> None:
    # Configuration
    lines_per_page = 4

    # Produce PostScript
    title = file.stem
    text = ("\n" * lines_per_page).join([str(i + 1) for i in range(pages)])
    subprocess.run(
        [
            "a2ps",
            f"--medium={paper}",
            f"--title={title}",
            f"--lines-per-page={lines_per_page}",
            "--portrait",
            "--columns=1",
            "--rows=1",
            f"--border={border}",
            "--no-header",
            f"--output={file.with_suffix('.ps')}",
        ],
        text=True,
        input=text,
        check=True,
    )

    # Convert to PDF if required
    if file.suffix == ".pdf":
        subprocess.check_call(
            [
                "ps2pdf",
                f"-sPAPERSIZE={paper}",
                f"{file.with_suffix('.ps')}",
                f"{file.with_suffix('.pdf')}",
            ]
        )
        os.remove(f"{file.with_suffix('.ps')}")



"""
extractres tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path

from pytest import mark, CaptureFixture

from testutils import file_test, compare_text_files, Case, GeneratedInput
from psutils.command.extractres import extractres

FIXTURE_DIR = Path(__file__).parent.resolve() / "test-files"


@mark.datafiles(
    FIXTURE_DIR / "extractres" / "sample" / "ISO-8859-1Encoding-expected.enc",
    FIXTURE_DIR / "extractres" / "sample" / "a2ps-a2ps-hdr2.02-expected.ps",
    FIXTURE_DIR / "extractres" / "sample" / "a2ps-black+white-Prolog2.01-expected.ps",
)

# pylint: disable=similarities
def test_extractres(
    capsys: CaptureFixture[str],
    datafiles: Path,
    regenerate_input: bool,
    regenerate_expected: bool,
) -> None:
    file_test(
        extractres,
        Case(
            "sample",
            [],
            GeneratedInput("a4", 1),
        ),
        FIXTURE_DIR,
        capsys,
        datafiles,
        ".ps",
        regenerate_input,
        regenerate_expected,
    )
    assert compare_text_files(
        capsys,
        datafiles / "ISO-8859-1Encoding.enc",
        datafiles / "ISO-8859-1Encoding-expected.enc",
    )
    assert compare_text_files(
        capsys,
        datafiles / "a2ps-a2ps-hdr2.02.ps",
        datafiles / "a2ps-a2ps-hdr2.02-expected.ps",
    )
    assert compare_text_files(
        capsys,
        datafiles / "a2ps-black+white-Prolog2.01.ps",
        datafiles / "a2ps-black+white-Prolog2.01-expected.ps",
    )



"""
includeres tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path
from typing import Callable, List

from pytest import mark, CaptureFixture

from testutils import file_test, make_tests, Case
from psutils.command.includeres import includeres

FIXTURE_DIR = Path(__file__).parent.resolve() / "test-files"

pytestmark = make_tests(
    includeres,
    FIXTURE_DIR,
    Case(
        "sample",
        [],
        str(FIXTURE_DIR / "extractres" / "sample" / "expected.ps"),
    ),
)


@mark.datafiles(
    FIXTURE_DIR / "includeres" / "sample" / "ISO-8859-1Encoding.enc",
    FIXTURE_DIR / "includeres" / "sample" / "a2ps-a2ps-hdr2.02.ps",
    FIXTURE_DIR / "includeres" / "sample" / "a2ps-black+white-Prolog2.01.ps",
)
# pylint: disable=similarities
def test_includeres(
    function: Callable[[List[str]], None],
    case: Case,
    fixture_dir: Path,
    capsys: CaptureFixture[str],
    datafiles: Path,
    regenerate_input: bool,
    regenerate_expected: bool,
) -> None:
    file_test(
        function,
        case,
        fixture_dir,
        capsys,
        datafiles,
        ".ps",
        regenerate_input,
        regenerate_expected,
    )



"""
psresize command.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import argparse
import sys
import warnings
from typing import List

from psutils.argparse import HelpFormatter, add_basic_arguments
from psutils.command.psnup import psnup
from psutils.warnings import simple_warning


def get_parser() -> argparse.ArgumentParser:
    # Command-line arguments
    parser = argparse.ArgumentParser(
        description="Change the page size of a PDF or PostScript document.",
        formatter_class=HelpFormatter,
        usage="%(prog)s [OPTION...] [INFILE [OUTFILE]]",
        add_help=False,
        epilog="""
PAGES is a comma-separated list of pages and page ranges; see
pstops(1) for more details.
    """,
    )
    warnings.showwarning = simple_warning(parser.prog)

    # Command-line parser
    parser.add_argument(
        "-p",
        "--paper",
        help="output paper name or dimensions (WIDTHxHEIGHT)",
    )
    parser.add_argument(
        "-P",
        "--inpaper",
        help="input paper name or dimensions (WIDTHxHEIGHT)",
    )
    add_basic_arguments(parser)

    # Backwards compatibility
    parser.add_argument("-w", "--width", help=argparse.SUPPRESS)
    parser.add_argument("-h", "--height", help=argparse.SUPPRESS)
    parser.add_argument("-W", "--inwidth", help=argparse.SUPPRESS)
    parser.add_argument("-H", "--inheight", help=argparse.SUPPRESS)

    return parser


# pylint: disable=dangerous-default-value
def psresize(argv: List[str] = sys.argv[1:]) -> None:
    args = get_parser().parse_intermixed_args(argv)

    # Resize pages
    cmd = ["-1"]
    if not args.verbose:
        cmd.append("--quiet")
    if args.width:
        cmd.extend(["--width", args.width])
    if args.height:
        cmd.extend(["--width", args.height])
    if args.inwidth:
        cmd.extend(["--inwidth", args.inwidth])
    if args.inheight:
        cmd.extend(["--inheight", args.inheight])
    if args.paper:
        cmd.extend(["--paper", args.paper])
    if args.inpaper:
        cmd.extend(["--inpaper", args.inpaper])
    if args.infile is not None:
        cmd.append(args.infile)
    if args.outfile is not None:
        cmd.append(args.outfile)
    psnup(cmd)


if __name__ == "__main__":
    psresize()



"""
psselect command.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import argparse
import sys
import warnings
from typing import List

from psutils.argparse import (
    HelpFormatter,
    PaperContext,
    add_basic_arguments,
    parserange,
    parsespecs,
)
from psutils.transformers import file_transform
from psutils.warnings import die, simple_warning


def get_parser() -> argparse.ArgumentParser:
    # Command-line arguments
    parser = argparse.ArgumentParser(
        description="Select pages from a PDF or PostScript document.",
        formatter_class=HelpFormatter,
        usage="%(prog)s [OPTION...] [PAGES] [INFILE [OUTFILE]]",
        add_help=False,
        epilog="""
PAGES is a comma-separated list of pages and page ranges; see
pstops(1) for more details.
""",
    )
    warnings.showwarning = simple_warning(parser.prog)

    # Command-line parser
    parser.add_argument("-R", "-p", "--pages", help="select the given page ranges")
    parser.add_argument(
        "-e", "--even", action="store_true", help="select even-numbered output pages"
    )
    parser.add_argument(
        "-o", "--odd", action="store_true", help="select odd-numbered output pages"
    )
    parser.add_argument(
        "-r",
        "--reverse",
        action="store_true",
        help="reverse the order of the output pages",
    )
    parser.add_argument("alt_pages", metavar="PAGES", nargs="?", help=argparse.SUPPRESS)
    add_basic_arguments(parser)

    return parser


# pylint: disable=dangerous-default-value
def psselect(argv: List[str] = sys.argv[1:]) -> None:
    args = get_parser().parse_intermixed_args(argv)

    # Get page range argument if supplied as a non-option
    if args.alt_pages is not None:
        if (
            args.pages is None
            and args.even is False
            and args.odd is False
            and args.reverse is False
        ):
            args.pages = args.alt_pages
        else:
            if args.outfile is not None:
                if args.pages is not None:
                    die("PAGES specified both with and without an option flag")
                else:
                    die("--pages must be used when --even, --odd or --reverse is used")
            args.outfile = args.infile
            args.infile = args.alt_pages

    # Rearrange pages
    pagerange = parserange(args.pages) if args.pages is not None else None
    paper_context = PaperContext()
    specs, modulo, flipping = parsespecs("0", paper_context)
    with file_transform(
        args.infile, args.outfile, None, None, specs, 0, False
    ) as transform:
        transform.transform_pages(
            pagerange, flipping, args.reverse, args.odd, args.even, modulo, args.verbose
        )


if __name__ == "__main__":
    psselect()



"""
setup.py for psutils
"""

import subprocess
from build_manpages.build_manpages import (  # type: ignore
    build_manpages,
    get_install_cmd,
    get_build_py_cmd,
)
from setuptools import setup

setup(
    cmdclass={
        "build_manpages": build_manpages,
        "build_py": get_build_py_cmd(),
        "install": get_install_cmd(),
    }
)

# Test we have 'paper' (from libpaper) installed
try:
    subprocess.check_output(["paper"])
except OSError as exc:
    raise SystemExit("psutils needs libpaper >= 2 to work") from exc



"""
psselect tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path

from testutils import file_test, make_tests, Case, GeneratedInput
from psutils.command.psselect import psselect

pytestmark = make_tests(
    psselect,
    Path(__file__).parent.resolve() / "test-files",
    Case(
        "odd",
        ["-o"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "even",
        ["-e"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "reverse",
        ["-r"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "even-reverse",
        ["-e", "-r"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "positive-range",
        ["--pages", "1-5"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "negative-range",
        ["-p", "_5-_1"],
        GeneratedInput("a4", 20),
    ),
    # Test psselect range going from positive to negative
    Case(
        "positive-negative-range",
        ["-R", "2-_2"],
        GeneratedInput("a4", 20),
    ),
    # Test psselect with short option and complex pagerange
    Case(
        "options-and-complex-pagerange",
        ["-o", "-p4-16,_3-_1"],
        GeneratedInput("a4", 20),
    ),
    # Test psselect with individual pages and ranges with -p
    Case(
        "individual-pages-and-dash-p",
        ["-p1,3,5,2,4,6,8-10,19"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "invalid-pagerange",
        ["-p", "1-5"],
        GeneratedInput("a4", 1),
        2,
    ),
    Case(
        "texlive",
        ["5-15"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "psnup-texlive",
        ["--pages", "1-18"],
        GeneratedInput("a4", 20),
    ),
)
test_psselect = file_test



"""
test_psnup tests.
Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

from pathlib import Path

from testutils import file_test, make_tests, Case, GeneratedInput
from psutils.command.psnup import psnup

pytestmark = make_tests(
    psnup,
    Path(__file__).parent.resolve() / "test-files",
    Case(
        "20-1",
        ["-p", "a4", "-1"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-2",
        ["-p", "a4", "-2"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-3",
        ["-p", "a4", "-3"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-4",
        ["-p", "a4", "-4"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-1-flip",
        ["-p", "a4", "-1", "-f"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-1-flipped-dimensions",
        ["-p", "297mmx210mm", "-1"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-1-flipped-dimensions-wrong-inpaper",
        ["-P", "297mmx210mm", "-p", "297mmx210mm", "-1"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-1-inpaper-A4-outpaper-A5",
        ["-p", "a5", "-1"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-1-inpaper-A5",
        ["-p", "a4", "-1"],
        GeneratedInput("a5", 20),
    ),
    Case(
        "20-1-inpaper-A5-set-inpaper",
        ["-P", "a5", "-p", "a4", "-1"],
        GeneratedInput("a5", 20),
    ),
    Case(
        "20-2-inpaper-A5",
        ["-p", "a4", "-2"],
        GeneratedInput("a5", 20),
    ),
    Case(
        "20-3-inpaper-A5",
        ["-p", "a4", "-3"],
        GeneratedInput("a5", 20),
    ),
    Case(
        "20-1-inpaper-A5-outpaper-A5",
        ["-p", "a5", "-1"],
        GeneratedInput("a5", 20),
    ),
    Case(
        "20-3-rotatedleft",
        ["-p", "a4", "-3", "-l"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-3-rotatedright",
        ["-p", "a4", "-3", "-r"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-3-impossible-tolerance",
        ["-p", "a4", "-3", "-t", "0"],
        GeneratedInput("a4", 1),
        1,
    ),
    Case(
        "20-4-flip",
        ["-p", "a4", "-4", "-f"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-4-border-20",
        ["-p", "a4", "-4", "-b", "20pt"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-4-margin-10",
        ["-p", "a4", "-4", "-m", "10pt"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-4-inpaper-A5",
        ["-p", "a4", "-4"],
        GeneratedInput("a5", 20),
    ),
    Case(
        "20-4-297mmx210mm",
        ["-p", "297mmx210mm", "-4"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-4-columnmajor",
        ["-p", "a4", "-4", "-c"],
        GeneratedInput("a4", 20),
    ),
    Case(
        "20-4-impossible-border",
        ["-p", "a4", "-4", "-b", "1000pt"],
        GeneratedInput("a4", 1),
        1,
    ),
    Case(
        "20-4-impossible-margin",
        ["-p", "a4", "-4", "-m", "1000pt"],
        GeneratedInput("a4", 1),
        1,
    ),
    Case(
        "draw",
        ["-p", "a4", "-4", "--draw=1pt"],
        GeneratedInput("a4", 4, 0),
    ),
    Case(
        "texlive",
        ["-pa4", "-2"],
        GeneratedInput("a4", 11),
    ),
    Case(
        "texlive2",
        ["-pa4", "-18"],
        "psselect-texlive-output",
    ),
    # The next 2 tests don't really apply to PDF, but run them anyway to
    # show that the warning is not generated for PDF.
    Case(
        "no-document-media",
        ["-2", "-ptabloid"],
        "no-document-media",
    ),
    Case(
        "no-document-media-explicit-size",
        ["-2", "-P612x792", "-ptabloid"],
        "no-document-media",
    ),
)
test_psnup = file_test
